#!/bin/sh

N_ACCESSES=$((3*1024))
# So that int32 accesses all work from 32k L1
MAX_IDX=$((64*1024 / 9))

# Another random, in case shell's RANDOM is too weak
rr=$$

gen_seq_walk() {
	i=$((N_ACCESSES / 3))
	while test $i != 0; do
		rr=$((rr * 75 % 65537))
		echo "a |= v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))];"
		rr=$((rr * 75 % 65537))
		echo "b |= v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))];"
		rr=$((rr * 75 % 65537))
		echo "c |= v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))];"
		: $((i--))
	done
}
gen_prog_walk() {
	echo "#include <stdint.h>"
	echo "int walk_$1(void *buffer)"
	echo "{"
	echo "$1 *v = ($1*)buffer;"
	echo "$1 a=0,b=0,c=0;"
	echo "$seq"
	echo "return a+b+c;"
	echo "}"
}

gen_seq_write() {
	i=$((N_ACCESSES / 3))
	while test $i != 0; do
		rr=$((rr * 75 % 65537))
		echo "v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))] = t;"
		rr=$((rr * 75 % 65537))
		echo "v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))] = t;"
		rr=$((rr * 75 % 65537))
		echo "v[$(( ((RANDOM + rr) & 0x7ffffff) % MAX_IDX))] = t;"
		: $((i--))
	done
}
gen_prog_write() {
	echo "#include <stdint.h>"
	echo "int write_$1(void *buffer)"
	echo "{"
	echo "$1 t = ($1)buffer;"
	echo "$1 *v = ($1*)buffer;"
	echo "$seq"
	echo "return 0;"
	echo "}"
}

echo "#define N_ACCESSES $N_ACCESSES" >N_ACCESSES.h
echo "#define MAX_IDX    $MAX_IDX" >>N_ACCESSES.h

# Note that the sequence of accesses is the same for all data types
seq=`gen_seq_walk`
gen_prog_walk int8_t  >"walk_int8.c"
gen_prog_walk int16_t >"walk_int16.c"
gen_prog_walk int32_t >"walk_int32.c"
gen_prog_walk int64_t >"walk_int64.c"

seq=`gen_seq_write`
gen_prog_write int8_t  >"write_int8.c"
gen_prog_write int16_t >"write_int16.c"
gen_prog_write int32_t >"write_int32.c"
gen_prog_write int64_t >"write_int64.c"
