#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "N_ACCESSES.h"

#define UNROLL_NUM 256

static unsigned time_us(void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_usec + tv.tv_sec * 1000000;
}

static long long b[1024*1024];

#if defined(__i386__) || defined(__x86_64__)
#define HAVE_RDTSC 1
static inline unsigned rdtsc(void)
{
	unsigned lo, hi;
	asm volatile("rdtsc" : "=a"(lo), "=d"(hi));
	return lo;
}
#endif
#ifdef HAVE_RDTSC
unsigned measure_rd(int (*f)(void*))
{
	unsigned mintime = ~0U;
	/* 200 was too small for CPU scaling to kick in on Core 2 Duo laptop */
	unsigned cnt = 1000;
	while (--cnt) {
		unsigned diff, beg;
		f(b); /* warm up */
		beg = rdtsc();
		f(b);
		diff = rdtsc() - beg;
		if (mintime > diff) {
			mintime = diff;
			//printf("mintime %u at cnt %u\n", mintime, cnt);
		}
	}
	return mintime;
}
#endif

unsigned measure_gt(int (*f)(void*))
{
	unsigned mintime = ~0U;
	unsigned cnt = 200;
	while (--cnt) {
		unsigned diff, beg;
		f(b);f(b); /* warm up */
		beg = time_us();
		/* 16x16 calls - must match UNROLL_NUM! */
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);f(b);
		diff = time_us() - beg;
		if (mintime > diff) {
			mintime = diff;
			//printf("mintime %u at cnt %u\n", mintime, cnt);
		}
	}
	return mintime;
}

static double mhz;

void measure_walk(int (*f)(void*), const char *name)
{
	/* On systems with dynamic CPU freq,
	 * measure_gt() speeds up CPU after a few loops,
	 * but measure_rd() may be too fast to do that.
	 * Thus call measure_gt() first:
	 */
	unsigned gt = measure_gt(f);
	printf("%s: %u us, %f ns/access", name,
			gt, (double)gt * 1000 / (UNROLL_NUM * N_ACCESSES)
	);
	if (mhz != 0.0) {
		printf(", %f clock/access",
			(double)gt * mhz / (UNROLL_NUM * N_ACCESSES)
		);
	}
#ifdef HAVE_RDTSC
	unsigned rd = measure_rd(f);
	printf(" (%u rdtsc ticks, %f/access)",
			rd, (double)rd/N_ACCESSES
	);
#endif
	printf("\n");
}

extern int walk_int8_t(void *buffer);
extern int walk_int16_t(void *buffer);
extern int walk_int32_t(void *buffer);
extern int walk_int64_t(void *buffer);

extern int write_int8_t(void *buffer);
extern int write_int16_t(void *buffer);
extern int write_int32_t(void *buffer);
extern int write_int64_t(void *buffer);

int main(int argc, char **argv)
{
	if (argv[1]) {
		mhz = atoi(argv[1]);
		printf("Assuming %.0f MHz CPU clock\n", mhz);
	} else {
		printf("CPU clock is not specified. It can be specified (in MHz) via 1st parameter\n");
	}

	printf("Each walk accesses %u*%u elements, max index is %u (x8=%u)\n",
			UNROLL_NUM, N_ACCESSES, MAX_IDX, MAX_IDX*8);
#ifdef HAVE_RDTSC
	printf("For rdtsc-based timing, each walk accesses %u elements\n", N_ACCESSES);
#endif

	measure_walk(walk_int8_t , "walk_int8_t");
	measure_walk(walk_int16_t, "walk_int16_t");
	measure_walk(walk_int32_t, "walk_int32_t");
	measure_walk(walk_int64_t, "walk_int64_t");

	measure_walk(write_int8_t , "write_int8_t");
	measure_walk(write_int16_t, "write_int16_t");
	measure_walk(write_int32_t, "write_int32_t");
	measure_walk(write_int64_t, "write_int64_t");

	return 0;
}
